class ShamirDecoder:
    def __init__(self, shares, prime):
        self.shares = shares
        self.prime = prime

    # Lagrange interpolation
    def _lagrange_interpolate(self, x, x_s, y_s):
        k = len(x_s)
        if k != len(set(x_s)):
            raise ValueError("points must be distinct")

        def PI(vals):
            accumulate = 1
            for v in vals:
                accumulate *= v
            return accumulate
        nums = []
        dens = []
        for i in range(k):
            others = list(x_s)
            cur = others.pop(i)
            nums.append(PI(x - o for o in others))
            dens.append(PI(cur - o for o in others))
        den = PI(dens)
        if den == 0:
            raise ValueError("zero denominator")
        num = sum(n * self._eval_at(dens[i], x, self.prime) * y_s[i]
                  for i, n in enumerate(nums))
        return (num * self._inv(den, self.prime)) % self.prime

    # Extended Euclidean Algorithm
    def _inv(self, a, b):
        t, new_t = 0, 1
        r, new_r = b, a
        while new_r != 0:
            quotient = r // new_r
            t, new_t = (new_t, t - quotient * new_t)
            r, new_r = (new_r, r - quotient * new_r)
        if r > 1:
            raise ValueError("a is not invertible")
        if t < 0:
            t += b
        return t

    # Evaluate polynomial
    def _eval_at(self, poly, x, prime):
        accumulate = 0
        for i in reversed(poly):
            accumulate *= x
            accumulate += i
            accumulate %= prime
        return accumulate

    # Recover secret
    def _recover_secret(self):
        x_s, y_s = zip(*self.shares)
        return self._lagrange_interpolate(0, x_s, y_s)

    def decode(self):
        return self._recover_secret()
